@@ -12,7 +12,90 @@ __docformat__ = 'restructuredtext'
 
 
 import numpy as N
-import nifticlib
+import nifti.clib as ncl
+
+#
+# type maps, codes and other NIfTI header stuff
+#
+filetypes = [ 'ANALYZE', 'NIFTI', 'NIFTI_PAIR', 'ANALYZE_GZ', 'NIFTI_GZ',
+              'NIFTI_PAIR_GZ' ]
+"""Typecodes of all supported NIfTI image formats."""
+
+N2nifti_dtype_map = { N.uint8: ncl.NIFTI_TYPE_UINT8,
+                      N.int8 : ncl.NIFTI_TYPE_INT8,
+                      N.uint16: ncl.NIFTI_TYPE_UINT16,
+                      N.int16 : ncl.NIFTI_TYPE_INT16,
+                      N.uint32: ncl.NIFTI_TYPE_UINT32,
+                      N.int32 : ncl.NIFTI_TYPE_INT32,
+                      N.uint64: ncl.NIFTI_TYPE_UINT64,
+                      N.int64 : ncl.NIFTI_TYPE_INT64,
+                      N.float32: ncl.NIFTI_TYPE_FLOAT32,
+                      N.float64: ncl.NIFTI_TYPE_FLOAT64,
+                      N.complex128: ncl.NIFTI_TYPE_COMPLEX128
+                    }
+"""Mapping of NumPy datatypes to NIfTI datatypes."""
+
+
+nifti2numpy_dtype_map = \
+    { ncl.NIFTI_TYPE_UINT8: 'u1',
+      ncl.NIFTI_TYPE_INT8: 'i1',
+      ncl.NIFTI_TYPE_UINT16: 'u2',
+      ncl.NIFTI_TYPE_INT16: 'i2',
+      ncl.NIFTI_TYPE_UINT32: 'u4',
+      ncl.NIFTI_TYPE_INT32: 'i4',
+      ncl.NIFTI_TYPE_UINT64: 'u8',
+      ncl.NIFTI_TYPE_INT64: 'i8',
+      ncl.NIFTI_TYPE_FLOAT32: 'f4',
+      ncl.NIFTI_TYPE_FLOAT64: 'f8',
+      ncl.NIFTI_TYPE_COMPLEX128: 'c16'
+    }
+"""Mapping of NIfTI to NumPy datatypes (necessary for handling memory
+mapped array with proper byte-order handling."""
+
+
+nifti_units_map = \
+    {"unknown": ncl.NIFTI_UNITS_UNKNOWN,
+     "m": ncl.NIFTI_UNITS_METER,
+     "mm": ncl.NIFTI_UNITS_MM,
+     "um": ncl.NIFTI_UNITS_MICRON,
+     "s": ncl.NIFTI_UNITS_SEC,
+     "ms": ncl.NIFTI_UNITS_MSEC,
+     "us": ncl.NIFTI_UNITS_USEC,
+     "Hz": ncl.NIFTI_UNITS_HZ,
+     "ppm": ncl.NIFTI_UNITS_PPM,
+     "rad/s": ncl.NIFTI_UNITS_RADS,
+    }
+
+
+# encode bits of NIfTI1 standard
+valid_xyz_unit_codes = range(8)
+valid_time_unit_codes = range(0, 64, 8)
+
+def Ndtype2niftidtype(array):
+    """Return the NIfTI datatype id for a corresponding NumPy datatype.
+    """
+    # get the real datatype from N type dictionary
+    dtype = N.typeDict[str(array.dtype)]
+
+    if not N2nifti_dtype_map.has_key(dtype):
+        raise ValueError, "Unsupported datatype '%s'" % str(array.dtype)
+
+    return N2nifti_dtype_map[dtype]
+
+
+nifti_xform_map = \
+    { 'unknown': ncl.NIFTI_XFORM_UNKNOWN,
+      'scanner': ncl.NIFTI_XFORM_SCANNER_ANAT,
+      'aligned': ncl.NIFTI_XFORM_ALIGNED_ANAT,
+      'talairach': ncl.NIFTI_XFORM_TALAIRACH,
+      'mni152': ncl.NIFTI_XFORM_MNI_152,
+    }
+nifti_xform_inv_map = dict([(v, k) for k, v in nifti_xform_map.iteritems()])
+
+
+#
+# utility functions
+#
 
 def time2vol( t, tr, lag=0.0, decimals=0 ):
     """ Translates a time 't' into a volume number. By default function returns
@@ -78,19 +161,23 @@ def applyFxToVolumes( ts, vols, fx, **kwargs ):
 
     return N.array( out )
 
+
 def getPeristimulusTimeseries( ts, onsetvols, nvols = 10, fx = N.mean ):
     """ Returns 4d array with peristimulus timeseries.
 
-    Parameters:
-        ts        - source 4d timeseries
-        onsetvols - sequence of onsetvolumes to be averaged over
-        nvols     - length of the peristimulus timeseries in volumes
-                    (starting from onsetvol)
-        fx        - function to be applied to the list of corresponding
-                    volumes. Typically this will be mean(), so it is default,
-                    but it could also be var() or something different. The
-                    supplied function is to be able to handle an 'axis=0'
-                    argument similiar to NumPy's mean(), var(), ...
+    :Parameters:
+      ts:
+        source 4d timeseries
+      onsetvols:
+        sequence of onsetvolumes to be averaged over
+      nvols:
+        length of the peristimulus timeseries in volumes (starting from
+        onsetvol)
+      fx:
+        function to be applied to the list of corresponding volumes. Typically
+        this will be mean(), so it is default, but it could also be var() or
+        something different. The supplied function is to be able to handle an
+        'axis=0' argument similiar to NumPy's mean(), var(), ...
     """
     selected = [ [ o + offset for o in onsetvols ] \
                     for offset in range( nvols ) ]
@@ -101,63 +188,10 @@ def getPeristimulusTimeseries( ts, onsetvols, nvols = 10, fx = N.mean ):
         return applyFxToVolumes( ts, selected, fx, axis=0 )
 
 
-filetypes = [ 'ANALYZE', 'NIFTI', 'NIFTI_PAIR', 'ANALYZE_GZ', 'NIFTI_GZ',
-              'NIFTI_PAIR_GZ' ]
-"""Typecodes of all supported NIfTI image formats."""
-
-N2nifti_dtype_map = { N.uint8: nifticlib.NIFTI_TYPE_UINT8,
-                      N.int8 : nifticlib.NIFTI_TYPE_INT8,
-                      N.uint16: nifticlib.NIFTI_TYPE_UINT16,
-                      N.int16 : nifticlib.NIFTI_TYPE_INT16,
-                      N.uint32: nifticlib.NIFTI_TYPE_UINT32,
-                      N.int32 : nifticlib.NIFTI_TYPE_INT32,
-                      N.uint64: nifticlib.NIFTI_TYPE_UINT64,
-                      N.int64 : nifticlib.NIFTI_TYPE_INT64,
-                      N.float32: nifticlib.NIFTI_TYPE_FLOAT32,
-                      N.float64: nifticlib.NIFTI_TYPE_FLOAT64,
-                      N.complex128: nifticlib.NIFTI_TYPE_COMPLEX128
-                    }
-"""Mapping of NumPy datatypes to NIfTI datatypes."""
-
-nifti2numpy_dtype_map = \
-    { nifticlib.NIFTI_TYPE_UINT8: 'u1',
-      nifticlib.NIFTI_TYPE_INT8: 'i1',
-      nifticlib.NIFTI_TYPE_UINT16: 'u2',
-      nifticlib.NIFTI_TYPE_INT16: 'i2',
-      nifticlib.NIFTI_TYPE_UINT32: 'u4',
-      nifticlib.NIFTI_TYPE_INT32: 'i4',
-      nifticlib.NIFTI_TYPE_UINT64: 'u8',
-      nifticlib.NIFTI_TYPE_INT64: 'i8',
-      nifticlib.NIFTI_TYPE_FLOAT32: 'f4',
-      nifticlib.NIFTI_TYPE_FLOAT64: 'f8',
-      nifticlib.NIFTI_TYPE_COMPLEX128: 'c16'
-    }
-"""Mapping of NIfTI to NumPy datatypes (necessary for handling memory
-mapped array with proper byte-order handling."""
-
-
-def Ndtype2niftidtype(array):
-    """Return the NIfTI datatype id for a corresponding NumPy datatype.
-    """
-    # get the real datatype from N type dictionary
-    dtype = N.typeDict[str(array.dtype)]
-
-    if not N2nifti_dtype_map.has_key(dtype):
-        raise ValueError, "Unsupported datatype '%s'" % str(array.dtype)
-
-    return N2nifti_dtype_map[dtype]
-
-
-nifti_xform_map = \
-    { 'unknown': nifticlib.NIFTI_XFORM_UNKNOWN,
-      'scanner': nifticlib.NIFTI_XFORM_SCANNER_ANAT,
-      'aligned': nifticlib.NIFTI_XFORM_ALIGNED_ANAT,
-      'talairach': nifticlib.NIFTI_XFORM_TALAIRACH,
-      'mni152': nifticlib.NIFTI_XFORM_MNI_152,
-    }
-
-
-def nhdr2dict(nhdr):
+#
+# little helpers
+#
+def nhdr2dict(nhdr, extensions=None):
     """Convert a NIfTI header struct into a python dictionary.
 
     While most elements of the header struct will be translated
@@ -166,6 +200,9 @@ def nhdr2dict(nhdr):
 
     :Parameters:
         nhdr: nifti_1_header
+        extensions: NiftiExtensions instance
+          All extensions will be merged into the returned dictionary
+          under the special `extensions` key.
 
     :Returns:
         dict
@@ -175,7 +212,7 @@ def nhdr2dict(nhdr):
     # the following header elements are converted in a simple loop
     # as they do not need special handling
     auto_convert = [ 'session_error', 'extents', 'sizeof_hdr',
-                     'slice_duration', 'slice_start', 'xyzt_units',
+                     'slice_duration', 'slice_start',
                      'cal_max', 'intent_p1', 'intent_p2', 'intent_p3',
                      'intent_code', 'sform_code', 'cal_min', 'scl_slope',
                      'slice_code', 'bitpix', 'descrip', 'glmin', 'dim_info',
@@ -191,17 +228,17 @@ def nhdr2dict(nhdr):
 
     # handle a few special cases
     # handle 'pixdim': carray -> list
-    pixdim = nifticlib.floatArray_frompointer(nhdr.pixdim)
+    pixdim = ncl.floatArray_frompointer(nhdr.pixdim)
     h['pixdim'] = [ pixdim[i] for i in range(8) ]
 
     # handle dim: carray -> list
-    dim = nifticlib.shortArray_frompointer(nhdr.dim)
+    dim = ncl.shortArray_frompointer(nhdr.dim)
     h['dim'] = [ dim[i] for i in range(8) ]
 
     # handle sform: carrays -> (4x4) ndarray
-    srow_x = nifticlib.floatArray_frompointer( nhdr.srow_x )
-    srow_y = nifticlib.floatArray_frompointer( nhdr.srow_y )
-    srow_z = nifticlib.floatArray_frompointer( nhdr.srow_z )
+    srow_x = ncl.floatArray_frompointer( nhdr.srow_x )
+    srow_y = ncl.floatArray_frompointer( nhdr.srow_y )
+    srow_z = ncl.floatArray_frompointer( nhdr.srow_z )
 
     h['sform'] = N.array( [ [ srow_x[i] for i in range(4) ],
                                 [ srow_y[i] for i in range(4) ],
@@ -212,6 +249,20 @@ def nhdr2dict(nhdr):
     h['quatern'] = [ nhdr.quatern_b, nhdr.quatern_c, nhdr.quatern_d ]
     h['qoffset'] = [ nhdr.qoffset_x, nhdr.qoffset_y, nhdr.qoffset_z ]
 
+    # some more postprocessing
+    # expand units
+    h['xyz_unit'] = ncl.xyzt2space(nhdr.xyzt_units)
+    h['time_unit'] = ncl.xyzt2time(nhdr.xyzt_units)
+
+    if not extensions:
+        return h
+
+    #
+    # handle extensions
+    #
+    # simply store a tuple of code (i.e. extension type) and extension data
+    h['extensions'] = [e for e in extensions.iteritems()]
+
     return h
 
 
@@ -265,7 +316,7 @@ def updateNiftiHeaderFromDict(nhdr, hdrdict):
         nhdr.dim_info = hdrdict['dim_info']
 
     if hdrdict.has_key('dim'):
-        dim = nifticlib.shortArray_frompointer(nhdr.dim)
+        dim = ncl.shortArray_frompointer(nhdr.dim)
         for i in range(8):
             dim[i] = hdrdict['dim'][i]
     if hdrdict.has_key('intent_p1'):
@@ -283,7 +334,7 @@ def updateNiftiHeaderFromDict(nhdr, hdrdict):
     if hdrdict.has_key('slice_start'):
         nhdr.slice_start = hdrdict['slice_start']
     if hdrdict.has_key('pixdim'):
-        pixdim = nifticlib.floatArray_frompointer(nhdr.pixdim)
+        pixdim = ncl.floatArray_frompointer(nhdr.pixdim)
         for i in range(8):
             pixdim[i] = hdrdict['pixdim'][i]
     if hdrdict.has_key('vox_offset'):
@@ -296,8 +347,21 @@ def updateNiftiHeaderFromDict(nhdr, hdrdict):
         nhdr.slice_end = hdrdict['slice_end']
     if hdrdict.has_key('slice_code'):
         nhdr.slice_code = hdrdict['slice_code']
-    if hdrdict.has_key('xyzt_units'):
-        nhdr.xyzt_units = hdrdict['xyzt_units']
+    if hdrdict.has_key('xyz_unit') \
+       or hdrdict.has_key('time_unit'):
+        # precharge units from current header, in case only one of them is to be
+        # updated
+        tu = ncl.xyzt2space(nhdr.xyzt_units)
+        su = ncl.xyzt2time(nhdr.xyzt_units)
+
+        # overwrite unit if present
+        if hdrdict.has_key('xyz_unit'):
+            su = _checkUnit(hdrdict['xyz_unit'], valid_xyz_unit_codes)
+        if hdrdict.has_key('time_unit'):
+            tu = _checkUnit(hdrdict['time_unit'], valid_time_unit_codes)
+        # compress both units into hdr format
+        nhdr.xyzt_units = ncl.spacetime2xyzt(su, tu)
+
     if hdrdict.has_key('cal_max'):
         nhdr.cal_max = hdrdict['cal_max']
     if hdrdict.has_key('cal_min'):
@@ -353,13 +417,13 @@ def updateNiftiHeaderFromDict(nhdr, hdrdict):
             raise ValueError, \
                   "Nifti header property 'sform' must be 4x4 matrix."
 
-        srow_x = nifticlib.floatArray_frompointer(nhdr.srow_x)
+        srow_x = ncl.floatArray_frompointer(nhdr.srow_x)
         for i in range(4):
             srow_x[i] = hdrdict['sform'][0][i]
-        srow_y = nifticlib.floatArray_frompointer(nhdr.srow_y)
+        srow_y = ncl.floatArray_frompointer(nhdr.srow_y)
         for i in range(4):
             srow_y[i] = hdrdict['sform'][1][i]
-        srow_z = nifticlib.floatArray_frompointer(nhdr.srow_z)
+        srow_z = ncl.floatArray_frompointer(nhdr.srow_z)
         for i in range(4):
             srow_z[i] = hdrdict['sform'][2][i]
 
@@ -385,6 +449,7 @@ def splitFilename(filename):
             Filename to be split.
 
     :Returns:
+      tuple: (basename, extension)
         The function returns a tuple of basename and extension. If no valid
         NIfTI filename extension is found, the whole string is returned as
         basename and the extension string will be empty.
@@ -401,3 +466,42 @@ def splitFilename(filename):
             return filename, ''
         else:
             return '.'.join(parts[:-1]), parts[-1]
+
+
+def _checkUnit(value, valid_codes):
+    """Internal helper function to check axis units and convert into codes.
+
+    :Parameter:
+      value: int | str
+      valid_codes: list
+
+    :Returns:
+      int
+
+    .. warning::
+      This is an internal method. Neither its availability nor its API is
+      guarenteed.
+    """
+    # is it a code?
+    if isinstance(value, int):
+        if value not in valid_codes:
+            raise ValueError, \
+                  "'%i' is not a valid NIfTI units code. Should be one of" \
+                  " %s." % (value,
+                            str([c for c in nifti_units_map.values()
+                                    if c in valid_codes]))
+        # value seems to be valid
+        return value
+
+    # must be plain unit then
+    else:
+        if value not in nifti_units_map.keys() \
+           or nifti_units_map[value] not in valid_codes:
+            raise ValueError, \
+                  "'%s' is not a valid NIfTI unit. Should be one of" \
+                  " %s." % (value,
+                            str([c for c in nifti_units_map.keys()
+                                    if nifti_units_map[c] in valid_codes]))
+
+        # value seems to be valid
+        return nifti_units_map[value]
